@@ -2,13 +2,15 @@ package Sphinx::Search;
 
 use warnings;
 use strict;
+
+use base 'Exporter';
+
 use Carp;
 use Socket;
 use Config;
-use Math::BigInt;
-use base 'Exporter';
-use Fcntl qw(:DEFAULT);
-use Errno;
+use Math::BigInt lib => 'GMP';
+use IO::Socket::INET;
+use IO::Socket::UNIX;
 
 =head1 NAME
 
@@ -18,11 +20,11 @@ Sphinx::Search - Sphinx search engine API Perl client
 
 Please note that you *MUST* install a version which is compatible with your version of Sphinx.
 
-This version is 0.12
+Use version 0.14 for Sphinx 0.9.9-rc1 and later
 
-Use version 0.12 for Sphinx 0.9.8 and later
+Use version 0.12 for Sphinx 0.9.8
 
-Use version 0.11 for Sphinx 0.9.8-rc1 and later
+Use version 0.11 for Sphinx 0.9.8-rc1
 
 Use version 0.10 for Sphinx 0.9.8-svn-r1112
 
@@ -38,7 +40,7 @@ Use version 0.02 for Sphinx 0.9.8-cvs-20070818
 
 =cut
 
-our $VERSION = '0.12';
+our $VERSION = '0.14';
 
 =head1 SYNOPSIS
 
@@ -73,11 +75,12 @@ use constant SEARCHD_COMMAND_SEARCH	=> 0;
 use constant SEARCHD_COMMAND_EXCERPT	=> 1;
 use constant SEARCHD_COMMAND_UPDATE	=> 2;
 use constant SEARCHD_COMMAND_KEYWORDS	=> 3;
+use constant SEARCHD_COMMAND_PERSIST	=> 4;
 
 # current client-side command implementation versions
-use constant VER_COMMAND_SEARCH		=> 0x113;
+use constant VER_COMMAND_SEARCH		=> 0x116;
 use constant VER_COMMAND_EXCERPT	=> 0x100;
-use constant VER_COMMAND_UPDATE	        => 0x101;
+use constant VER_COMMAND_UPDATE	        => 0x102;
 use constant VER_COMMAND_KEYWORDS       => 0x100;
 
 # known searchd status codes
@@ -100,6 +103,8 @@ use constant SPH_RANK_PROXIMITY_BM25    => 0; # default mode, phrase proximity m
 use constant SPH_RANK_BM25              => 1; # statistical mode, BM25 ranking only (faster but worse quality)
 use constant SPH_RANK_NONE              => 2; # no ranking, all matches get a weight of 1
 use constant SPH_RANK_WORDCOUNT         => 3; # simple word-count weighting, rank is a weighted sum of per-field keyword occurence counts
+use constant SPH_RANK_PROXIMITY         => 4;
+use constant SPH_RANK_MATCHANY          => 5;
 
 # known sort modes
 use constant SPH_SORT_RELEVANCE		=> 0;
@@ -120,6 +125,7 @@ use constant SPH_ATTR_TIMESTAMP		=> 2;
 use constant SPH_ATTR_ORDINAL		=> 3;
 use constant SPH_ATTR_BOOL		=> 4;
 use constant SPH_ATTR_FLOAT		=> 5;
+use constant SPH_ATTR_BIGINT		=> 6;
 use constant SPH_ATTR_MULTI		=> 0x40000000;
 
 # known grouping functions
@@ -133,6 +139,103 @@ use constant SPH_GROUPBY_ATTRPAIR	=> 5;
 # Floating point number matching expression
 my $num_re = qr/^-?\d*\.?\d*(?:[eE][+-]?\d+)?$/;
 
+# portably pack numeric to 64 signed bits, network order
+sub _sphPackI64 {
+    my $self = shift;
+    my $v = shift;
+
+    # x64 route
+    if ( $self->{_longsize} >= 8 ) {
+	my $i = int($v);
+	return pack ( "NN", $i>>32, $i & 4294967295 );
+    }
+
+    # x32 route, BigInt
+    my $vb = Math::BigInt->new($v);
+    my $l = $vb->copy->band ( 4294967295 );
+    my $h = $vb->copy->brsft ( 32 );
+    return pack ( "NN", $h, $l );
+}
+
+# portably pack numeric to 64 unsigned bits, network order
+sub _sphPackU64 {
+    my $self = shift;
+    my $v = shift;
+
+    # x64 route
+    if ( $self->{_longsize} >= 8 ) {
+	my $i = int($v);
+	return pack ( "NN", $i>>32, $i & 4294967295 );
+    }
+
+    # x32 route, BigInt
+    my $vb = Math::BigInt->new($v);
+    my $l = $vb->copy->band ( 4294967295 );
+    my $h = $vb->copy->brsft ( 32 );
+    return pack ( "NN", $h, $l );
+}
+
+sub _sphPackI64array {
+    my $self = shift;
+    my $values = shift || [];
+
+    my $s = pack("N", scalar @$values);
+    $s .= $self->_sphPackI64($_) for @$values;
+    return $s;
+}
+
+# portably unpack 64 unsigned bits, network order to numeric
+sub _sphUnpackU64 
+{
+    my $self = shift;
+    my $v = shift;
+
+    my ($h,$l) = unpack ( "N*N*", $v );
+
+    # x64 route
+    return ($h<<32) + $l if ( $self->{_longsize} >= 8 );
+
+    # x32 route, BigInt
+    $h = Math::BigInt->new($h);
+    $h->blsft(32)->badd($l);
+    
+    return $h->bstr;
+}
+
+# portably unpack 64 signed bits, network order to numeric
+sub _sphUnpackI64 
+{
+    my $self = shift;
+    my $v = shift;
+
+    my ($h,$l) = unpack ( "N*N*", $v );
+
+    my $neg = ($h & 0x80000000) ? 1 : 0;
+
+    # x64 route
+    if ( $self->{_longsize} >= 8 ) {
+	return -(~(($h<<32) + $l) + 1) if $neg;
+	return ($h<<32) + $l;
+    }
+
+    # x32 route, BigInt
+    if ($neg) {
+	$h = ~$h;
+	$l = ~$l;
+    }
+
+    my $x = Math::BigInt->new($h);
+    $x->blsft(32)->badd($l);
+    $x->binc()->bneg() if $neg;
+
+    return $x->bstr;
+}
+
+
+
+
+
+
 =head1 CONSTRUCTOR
 
 =head2 new
@@ -168,6 +271,8 @@ sub new {
 	# per=client-object settings
 	_host		=> 'localhost',
 	_port		=> 3312,
+	_path           => undef,
+	_socket         => undef,
 
 	# per-query settings
 	_offset		=> 0,
@@ -192,17 +297,18 @@ sub new {
 	_ranker         => SPH_RANK_PROXIMITY_BM25,
 	_maxquerytime   => 0,
 	_fieldweights   => {},
+	_overrides      => {},
+	_select         => q{*},
 
 	# per-reply fields (for single-query case)
 	_error		=> '',
 	_warning	=> '',
+	_connerror      => '',,
 	
 	# request storage (for multi-query case)
 	_reqs           => [],
-
 	_timeout        => 0,
-	_connection     => undef,
-	_persistent_connection => 0,
+
 	_longsize      => $Config{longsize},
     };
     bless $self, ref($class) || $class;
@@ -219,6 +325,7 @@ sub new {
     return $self;
 }
 
+
 =head1 METHODS
 
 =cut
@@ -263,133 +370,141 @@ sub GetLastWarning {
 	return $self->{_warning};
 }
 
-# Persistent connections not currently supported by searchd.
-#=head2 SetPersistentConnection
-#
-#    $sphinx->SetPersistentConnection(1);
-#    $sphinx->SetPersistentConnection(0);
-#
-#Enable/disable persistent connections.
-#
-#When enabled, the connection to the searchd server is kept open; you must close
-#it manually using Disconnect when you have finished with it.  This is the most
-#efficient way to reach the search server if you are performing multiple queries
-#with the one Sphinx::Search instance, such as in a web application server
-#environment.
-#
-#When disabled, the connection is established and automatically disconnected on
-#each query.  This is useful for CGI type environments where you have only one
-#request for the lifetime of the process.
-#
-#=cut
-#
-sub _SetPersistentConnection {
+
+=head2 IsConnectError 
+
+Check connection error flag (to differentiate between network connection errors
+and bad responses).  Returns true value on connection error.
+
+=cut
+
+sub IsConnectError {
+    return shift->{_connerror};
+}
+
+=head2 SetServer
+
+    $sph->SetServer($host, $port);
+    $sph->SetServer($path, $port);
+
+In the first form, sets the host (string) and port (integer) details for the
+searchd server using a network (INET) socket.
+
+In the second form, where $path is a local filesystem path (optionally prefixed
+by 'unix://'), sets the client to access the searchd server via a local (UNIX
+domain) socket at the specified path.
+
+Returns $sph.
+
+=cut
+
+sub SetServer {
     my $self = shift;
-    $self->{_persistent_connection} = shift;
+    my $host = shift;
+    my $port = shift;
+
+    croak("host is not defined") unless defined($host);
+    $self->{_path} = $host, return if substr($host, 0, 1) eq '/';
+    $self->{_path} = substr($host, 7), return if substr($host, 0, 7) eq 'unix://';
+	
+    croak("port is not an integer") unless defined($port) && $port =~ m/^\d+$/o;
+
+    $self->{_host} = $host;
+    $self->{_port} = $port;
+    $self->{_path} = undef;
+
     return $self;
 }
-#
-#=head2 Disconnect
-#
-#    $sph->Disconnect;
-#
-#Disconnect from the search server.  See L<SetPersistentConnection> for a description.
-#
-#=cut
-#
-sub _Disconnect {
-    my $self = shift;
 
-    if ($self->{_connection}) {
-	close($self->{_connection});
-	$self->{_connection} = undef;
+=head2 SetConnectTimeout
+
+    $sph->SetConnectTimeout($timeout)
+
+Set server connection timeout (in seconds).
+
+Returns $sph.
+
+=cut
+
+sub SetConnectTimeout {
+    my $self = shift;
+    my $timeout = shift;
+
+    croak("timeout is not numeric") unless ($timeout =~  m/$num_re/);
+    $self->{_timeout} = $timeout;
+}
+
+sub _Send {
+    my $self = shift;
+    my $fp = shift;
+    my $data = shift;
+
+    $self->{_log}->debug("Writing to socket") if $self->{_debug};
+    $fp->write($data); return 1;
+    if ($fp->eof || ! $fp->write($data)) {
+	$self->_Error("connection unexpectedly closed (timed out?): $!");
+	$self->{_connerror} = 1;
+	return 0;
     }
-    return $self;
+    return 1;
 }
 
-
-#-------------------------------------------------------------
-
 # connect to searchd server
 
 sub _Connect {
 	my $self = shift;
 	
+	return $self->{_socket} if $self->{_socket};
+
 	my $debug = $self->{_debug};
-
-	$self->{_log}->debug("Connecting to $self->{_host}:$self->{_port}") if $debug;
-
-	return $self->{_connection} if $self->{_connection};
+	my $str_dest = $self->{_path} ? 'unix://' . $self->{_path} : "$self->{_host}:$self->{_port}";
+	$self->{_log}->debug("Connecting to $str_dest") if $debug;
 
 	# connect socket
-	my $fp;
-	socket($fp, PF_INET, SOCK_STREAM, getprotobyname('tcp')) || Carp::croak("socket: ".$!);
-	binmode($fp, ':bytes');
-	my $dest = sockaddr_in($self->{_port}, inet_aton($self->{_host}));
+	$self->{_connerror} = q{};
 
-	if ($self->{_timeout}) {
-	    my $flags = fcntl($fp, F_GETFL, 0) or do {
-		$self->_Error("Can't get flags for socket: $!");
-		return 0;
-	    };
-	    fcntl($fp, F_SETFL, $flags | O_NONBLOCK) or do {
-		$self->_Error("Can't set flags for socket: $!");
-		return 0;
-	    };
-		
-	    if (! connect($fp, $dest)) {
-		if (! $!{EINPROGRESS}) {
-		    $self->_Error("connect to {$self->{_host}}:{$self->{_port}} failed: $!");
-		    return 0;
-		}
-		my $count = 0;
-		while ($count < 2) { # until timeout or data received
-		    my $vec = '';
-		    vec($vec, fileno($fp), 1) = 1;
-		    my $n = select($vec, undef, undef, $self->{_timeout});
-		    if ($n < 0) {
-			$self->_Error("connection to {$self->{_host}}:{$self->{_port}} failed: $!");
-			return 0;
-		    }
-		    elsif ($n == 0) {
-			$self->_Error("connection to {$self->{_host}}:{$self->{_port}} timed out");
-			return 0;
-		    }
-		    else {
-			my $tmpbuf = 0;
-			my $ret = recv($fp, $tmpbuf, 4, MSG_PEEK);
-			last if $tmpbuf;
-		    }
-		    $count++;
-		}
-	    }
-	    fcntl($fp, F_SETFL, $flags) or do {
-		$self->_Error("Can't set flags for socket: $!");
-		return 0;
-	    };
+	my $fp;
+	my %params = (); # ( Blocking => 0 );
+	$params{Timeout} = $self->{_timeout} if $self->{_timeout};
+	if ($self->{_path}) {
+	    $fp = IO::Socket::UNIX->new( Peer => $self->{_path},
+					 %params,
+					 );
 	}
 	else {
-	    connect($fp, $dest) or do {
-		$self->_Error("connection to {$self->{_host}}:{$self->{_port}} failed: $!");
-		return 0;
-	    };
+	    $fp = IO::Socket::INET->new( PeerPort => $self->{_port},
+					 PeerAddr => $self->{_host},
+					 Proto => 'tcp',
+					 %params,
+					 );
 	}
+	if (! $fp) {
+	    $self->_Error("Failed to open connection to $str_dest: $!");
+	    $self->{_connerror} = 1;
+	    return 0;
+	}
+	binmode($fp, ':bytes');
+
 	# check version
 	my $buf = '';
-	croak("recv: ".$!) unless defined recv($fp, $buf, 4, 0);
+	$fp->read($buf, 4) or do {
+	    $self->_Error("Failed on initial read from $str_dest: $!");
+	    $self->{_connerror} = 1;
+	    return 0;
+	};
 	my $v = unpack("N*", $buf);
 	$v = int($v);
-	if($v < 1) {
-		close($fp) || croak("close: $!");
-		$self->_Error("expected searchd protocol version 1+, got version '$v'");
+	$self->{_log}->debug("Got version $v from searchd") if $debug;
+	if ($v < 1) {
+	    close($fp);
+	    $self->_Error("expected searchd protocol version 1+, got version '$v'");
+	    return 0;
 	}
 
 	$self->{_log}->debug("Sending version") if $debug;
 
 	# All ok, send my version
-	send($fp, pack("N", 1),0);
-
-	$self->{_connection} = $fp;
+	$self->_Send($fp, pack("N", 1)) or return 0;
 
 	$self->{_log}->debug("Connection complete") if $debug;
 
@@ -405,19 +520,27 @@ sub _GetResponse {
 	my $client_ver = shift;
 
 	my $header;
-	croak("recv: ".$!) unless defined recv($fp, $header, 8, 0);
+	defined($fp->read($header, 8, 0)) or do {
+	    $self->_Error("read failed: $!");
+	    return 0;
+	};
 
 	my ($status, $ver, $len ) = unpack("n2N", $header);
-        my ($chunk, $response);
-	while(defined($chunk = <$fp>)) {
-		$response .= $chunk;
+        my $response = q{};
+	my $lasterror = q{};
+	my $lentotal = 0;
+	while (my $rlen = $fp->read(my $chunk, $len)) {
+	    $lasterror = $!, last if $rlen < 0;
+	    $response .= $chunk;
+	    $lentotal += $rlen;
+	    last if $lentotal >= $len;
 	}
-        $self->_Disconnect unless $self->{_persistent_connection};
+        close($fp) unless $self->{_socket};
 
 	# check response
-        if ( !$response || length($response) != $len ) {
+        if ( length($response) != $len ) {
 		$self->_Error( $len 
-			? "failed to read searchd response (status=$status, ver=$ver, len=$len, read=". length($response) . ")"
+			? "failed to read searchd response (status=$status, ver=$ver, len=$len, read=". length($response) . ", last error=$lasterror)"
        			: "received zero-sized searchd response");
 		return 0;
 	}
@@ -449,85 +572,6 @@ sub _GetResponse {
         return $response;
 }
 
-# portably pack numeric to 64 unsigned bits, network order
-sub _sphPack64 {
-    my $self = shift;
-    my $v = shift;
-
-    # x64 route
-    if ( $self->{_longsize} >= 8 ) {
-	my $i = int($v);
-	return pack ( "NN", $i>>32, $i & 4294967295 );
-    }
-
-    # x32 route, BigInt
-    my $x = "4294967296";
-    my $vb = Math::BigInt->new($v);
-    my $l = $vb->copy->band ( 4294967295 );
-    my $h = $vb->copy->brsft ( 32 );
-    return pack ( "NN", $h, $l );
-}
-
-
-# portably unpack 64 unsigned bits, network order to numeric
-sub _sphUnpack64 
-{
-    my $self = shift;
-    my $v = shift;
-
-    my ($h,$l) = unpack ( "N*N*", $v );
-
-    # x64 route
-    return ($h<<32) + $l if ( $self->{_longsize} >= 8 );
-
-    # x32 route, BigInt
-    $h = Math::BigInt->new($h);
-    $h->blsft(32)->badd($l);
-    
-    return $h->bstr;
-}
-
-
-
-=head2 SetServer
-
-    $sph->SetServer($host, $port);
-
-Set the host (string) and port (integer) details for the searchd server.  Returns $sph.
-
-=cut
-
-sub SetServer {
-    my $self = shift;
-    my $host = shift;
-    my $port = shift;
-
-    croak("host is not defined") unless defined($host);
-    croak("port is not an integer") unless defined($port) && $port =~ m/^\d+$/o;
-
-    $self->{_host} = $host;
-    $self->{_port} = $port;
-    return $self;
-}
-
-=head2 SetConnectTimeout
-
-    $sph->SetConnectTimeout($timeout)
-
-Set server connection timeout (in seconds).
-
-Returns $sph.
-
-=cut
-
-sub SetConnectTimeout {
-    my $self = shift;
-    my $timeout = shift;
-
-    croak("timeout is not numeric") unless ($timeout =~  m/$num_re/);
-    $self->{_timeout} = $timeout;
-}
-
 =head2 SetLimits
 
     $sph->SetLimits($offset, $limit);
@@ -665,7 +709,8 @@ sub SetRankingMode {
     croak("Unknown ranking mode: $ranker") unless ( $ranker==SPH_RANK_PROXIMITY_BM25
 						    || $ranker==SPH_RANK_BM25
 						    || $ranker==SPH_RANK_NONE
-						    || $ranker==SPH_RANK_WORDCOUNT );
+						    || $ranker==SPH_RANK_WORDCOUNT
+						    || $ranker==SPH_RANK_PROXIMITY );
 
     $self->{_ranker} = $ranker;
     return $self;
@@ -865,8 +910,6 @@ Sets the results to be filtered on a range of values for the given
 attribute. Only those records where $attr column value is between $min and $max
 (including $min and $max) will be returned.
 
-$min and $max must be integers.  Use L<SetFilterFloatRange> for floating point values.
-
 If 'exclude' is set, excludes results that fall within the given range.
 
 Returns $sph.
@@ -876,8 +919,8 @@ Returns $sph.
 sub SetFilterRange {
     my ($self, $attribute, $min, $max, $exclude) = @_;
     croak("attribute is not defined") unless (defined $attribute);
-    croak("min: $min is not an integer") unless ($min =~ /^-?\d+$/);
-    croak("max: $max is not an integer") unless ($max =~ /^-?\d+$/);
+    croak("min: $min is not an integer") unless ($min =~ m/$num_re/);
+    croak("max: $max is not an integer") unless ($max =~ m/$num_re/);
     croak("min value should be <= max") unless ($min <= $max);
 
     push(@{$self->{_filters}}, {
@@ -1098,6 +1141,49 @@ sub SetRetries {
     return $self;
 }
 
+=head2 SetOverride
+
+    $sph->SetOverride($attrname, $attrtype, $values);
+
+ Set attribute values override. There can be only one override per attribute.
+ $values must be a hash that maps document IDs to attribute values
+
+=cut
+
+sub SetOverride {
+    my $self = shift;
+    my $attrname = shift;
+    my $attrtype = shift;
+    my $values = shift;
+
+    croak("attribute name is not defined") unless defined $attrname;
+    croak("Uknown attribute type: $attrtype") unless ($attrtype == SPH_ATTR_INTEGER
+						      || $attrtype == SPH_ATTR_TIMESTAMP
+						      || $attrtype == SPH_ATTR_BOOL
+						      || $attrtype == SPH_ATTR_FLOAT
+						      || $attrtype == SPH_ATTR_BIGINT);
+    $self->{_overrides}->{$attrname} = { attr => $attrname,
+					 type => $attrtype,
+					 values => $values,
+				     };
+    
+    return $self;
+}
+
+
+=head2 SetSelect 
+
+    $sph->SetSelect($select)
+
+Set select list (attributes or expressions).  SQL-like syntax.
+
+=cut
+
+sub SetSelect {
+    my $self = shift;
+    $self->{_select} = shift;
+    return $self;
+}
 
 =head2 ResetFilters
 
@@ -1135,6 +1221,18 @@ sub ResetGroupBy {
     return $self;
 }
 
+=head2 ResetOverrides
+
+Clear all attribute value overrides (for multi-queries)
+
+=cut
+
+sub ResetOverrides {
+    my $self = shift;
+
+    $self->{_select} = undef;
+    return $self;
+}
 
 =head2 Query
 
@@ -1245,8 +1343,8 @@ sub AddQuery {
     $req .= pack ( "N*", scalar(@{$self->{_weights}}), @{$self->{_weights}});
     $req .= pack ( "N/a*", $index); # indexes
     $req .= pack ( "N", 1) 
-	. $self->_sphPack64($self->{_min_id})
-	. $self->_sphPack64($self->{_max_id}); # id64 range
+	. $self->_sphPackU64($self->{_min_id})
+	. $self->_sphPackU64($self->{_max_id}); # id64 range
 
     # filters
     $req .= pack ( "N", scalar @{$self->{_filters}} );
@@ -1256,10 +1354,10 @@ sub AddQuery {
 
 	my $t = $filter->{type};
 	if ($t == SPH_FILTER_VALUES) {
-	    $req .= pack ( "N*", scalar(@{$filter->{values}}), @{$filter->{values}});
+	    $req .= $self->_sphPackI64array($filter->{values});
 	}
 	elsif ($t == SPH_FILTER_RANGE) {
-	    $req .= pack ( "NN", $filter->{min}, $filter->{max} );
+	    $req .= $self->_sphPackI64($filter->{min}) . $self->_sphPackI64($filter->{max});
 	}
 	elsif ($t == SPH_FILTER_FLOATRANGE) {
 	    $req .= _PackFloat ( $filter->{"min"} ) . _PackFloat ( $filter->{"max"} );
@@ -1301,6 +1399,31 @@ sub AddQuery {
     # comment
     $req .= pack ( "N/a*", $comment);
 
+    # attribute overrides
+    $req .= pack ( "N", scalar keys %{$self->{_overrides}} );
+    for my $entry (values %{$self->{_overrides}}) {
+	$req .= pack ("N/a*", $entry->{attr})
+	    . pack ("NN", $entry->{type}, scalar keys %{$entry->{values}});
+	for my $id (keys %{$entry->{values}}) {
+	    croak "Attribute value key is not numeric" unless $id =~ m/$num_re/;
+	    my $v = $entry->{values}->{$id};
+	    croak "Attribute value key is not numeric" unless $v =~ m/$num_re/;
+	    $req .= $self->_sphPackU64($id);
+	    if ($entry->{type} == SPH_ATTR_FLOAT) {
+		$req .= $self->_packfloat($v);
+	    }
+	    elsif ($entry->{type} == SPH_ATTR_BIGINT) {
+		$req .= $self->_sphPackI64($v);
+	    }
+	    else {
+		$req .= pack("N", $v);
+	    }
+	}
+    }
+    
+    # select list
+    $req .= pack("N/a*", $self->{_select} || '');
+
     push(@{$self->{_reqs}}, $req);
 
     return scalar $#{$self->{_reqs}};
@@ -1349,7 +1472,7 @@ sub RunQueries {
     my $nreqs = @{$self->{_reqs}};
     my $req = pack("Na*", $nreqs, join("", @{$self->{_reqs}}));
     $req = pack ( "nnN/a*", SEARCHD_COMMAND_SEARCH, VER_COMMAND_SEARCH, $req); # add header
-    send($fp, $req ,0);
+    $self->_Send($fp, $req);
 
     $self->{_reqs} = [];
 		   
@@ -1414,7 +1537,7 @@ sub RunQueries {
 	while ( $count-->0 && $p<$max ) {
 	    my $data = {};
 	    if ($id64) {
-		$data->{doc} = $self->_sphUnpack64(substr($response, $p, 8)); $p += 8;
+		$data->{doc} = $self->_sphUnpackU64(substr($response, $p, 8)); $p += 8;
 		$data->{weight} = unpack("N*", substr($response, $p, 4)); $p += 4;
 	    }
 	    else {
@@ -1422,6 +1545,10 @@ sub RunQueries {
 		$p += 8;
 	    }
 	    foreach my $attr (@attr_list) {
+		if ($attrs{$attr} == SPH_ATTR_BIGINT) {
+		    $data->{$attr} = $self->_sphUnpackI64(substr($response, $p, 8)); $p += 8;
+		    next;
+		}
 		if ($attrs{$attr} == SPH_ATTR_FLOAT) {
 		    my $uval = unpack( "N*", substr ( $response, $p, 4 ) ); $p += 4;
 		    $data->{$attr} = [ unpack("f*", pack("L", $uval)) ];
@@ -1578,7 +1705,7 @@ sub BuildExcerpts {
 	##########################
 
 	$req = pack ( "nnN/a*", SEARCHD_COMMAND_EXCERPT, VER_COMMAND_EXCERPT, $req); # add header
-	send($fp, $req ,0);
+	$self->_Send($fp, $req);
 	
 	my $response = $self->_GetResponse($fp, VER_COMMAND_EXCERPT);
 	return 0 unless ($response);
@@ -1646,7 +1773,7 @@ sub BuildKeywords {
     ##################
 
     $req = pack ( "nnN/a*", SEARCHD_COMMAND_KEYWORDS, VER_COMMAND_KEYWORDS, $req);
-    send($fp, $req, 0);
+    $self->_Send($fp, $req);
     my $response = $self->_GetResponse ( $fp, VER_COMMAND_KEYWORDS );
     return unless $response;
 
@@ -1701,6 +1828,7 @@ sub EscapeString {
 =head2 UpdateAttributes
 
     $sph->UpdateAttributes($index, \@attrs, \%values);
+    $sph->UpdateAttributes($index, \@attrs, \%values, $mva);
 
 Update specified attributes on specified documents
 
@@ -1731,7 +1859,7 @@ Usage example:
 =cut
 
 sub UpdateAttributes  {
-    my ($self, $index, $attrs, $values ) = @_;
+    my ($self, $index, $attrs, $values, $mva ) = @_;
 
     croak("index is not defined") unless (defined $index);
     croak("attrs must be an array") unless ref($attrs) eq "ARRAY";
@@ -1746,7 +1874,15 @@ sub UpdateAttributes  {
 	croak("value entry must be an array") unless ref($entry) eq "ARRAY";
 	croak("size of values must match size of attrs") unless @$entry == @$attrs;
 	for my $v (@$entry) {
-	    croak("entry value $v is not an integer") unless ($v =~ /^(\d+)$/o);
+	    if ($mva) {
+		croak("multi-valued entry $v is not an array") unless ref($v) eq 'ARRAY';
+		for my $vv (@$v) {
+		    croak("array entry value $vv is not an integer") unless ($vv =~ /^(\d+)$/o);
+		}
+	    }
+	    else { 
+		croak("entry value $v is not an integer") unless ($v =~ /^(\d+)$/o);
+	    }
 	}
     }
 
@@ -1755,14 +1891,25 @@ sub UpdateAttributes  {
 
     $req .= pack ( "N", scalar @$attrs );
     for my $attr (@$attrs) {
-	$req .= pack ( "N/a*", $attr);
+	$req .= pack ( "N/a*", $attr)
+	    . pack("N", $mva ? 1 : 0);
     }
     $req .= pack ( "N", scalar keys %$values );
     foreach my $id (keys %$values) {
 	my $entry = $values->{$id};
-	$req .= $self->_sphPack64($id);
-	for my $v ( @$entry ) {
-	    $req .= pack ( "N", $v );
+	$req .= $self->_sphPackU64($id);
+	if ($mva) {
+	    for my $v ( @$entry ) {
+		$req .= pack ( "N", @$v );
+		for my $vv (@$v) {
+		    $req .= pack ("N", $vv);
+		}
+	    }
+	}
+	else {
+	    for my $v ( @$entry ) {
+		$req .= pack ( "N", $v );
+	    }
 	}
     }
 
@@ -1780,6 +1927,61 @@ sub UpdateAttributes  {
     return $updated;
 }
 
+=head2 Open
+
+    $sph->Open()
+
+Opens a persistent connection for subsequent queries.  
+
+To reduce the network connection overhead of making Sphinx queries, you can call
+$sph->Open(), then run any number of queries, and call $sph->Close() when
+finished.
+
+Returns 1 on success, 0 on failure.
+
+=cut 
+
+sub Open {
+    my $self = shift;
+
+    if ($self->{_socket}) {
+	$self->_Error("already connected");
+	return 0;
+    }
+    my $fp = $self->_Connect() or return 0;
+
+    my $req = pack("nnNN", SEARCHD_COMMAND_PERSIST, 0, 4, 1);
+    $self->_Send($fp, $req) or return 0;
+
+    $self->{_socket} = $fp;
+    return 1;
+}
+
+=head2 Close
+
+    $sph->Close()
+
+Closes a persistent connection.
+
+Returns 1 on success, 0 on failure.
+
+=cut 
+
+sub Close {
+    my $self = shift;
+
+    if (! $self->{_socket}) {
+	$self->_Error("not connected");
+	return 0;
+    }
+    
+    close($self->{_socket});
+    $self->{_socket} = undef;
+
+    return 1;
+}
+
+
 =head1 SEE ALSO
 
 L<http://www.sphinxsearch.com>
